@@ -58,7 +58,7 @@ void make_parts(char *dev, int cnt, int symlinks)
 	 * else that of dev
 	 */
 	struct stat stb;
-	int major, minor;
+	int major_num, minor_num;
 	int i;
 	int nlen = strlen(dev) + 20;
 	char *name = malloc(nlen);
@@ -69,21 +69,21 @@ void make_parts(char *dev, int cnt, int symlinks)
 		return;
 	if (!S_ISBLK(stb.st_mode))
 		return;
-	major = major(stb.st_rdev);
-	minor = minor(stb.st_rdev);
+	major_num = major(stb.st_rdev);
+	minor_num = minor(stb.st_rdev);
 	for (i=1; i <= cnt ; i++) {
 		struct stat stb2;
 		snprintf(name, nlen, "%s%s%d", dev, dig?"p":"", i);
 		if (stat(name, &stb2)==0) {
 			if (!S_ISBLK(stb2.st_mode))
 				continue;
-			if (stb2.st_rdev == makedev(major, minor+i))
+			if (stb2.st_rdev == makedev(major_num, minor_num+i))
 				continue;
 			unlink(name);
 		} else {
 			stb2 = stb;
 		}
-		if (mknod(name, S_IFBLK | 0600, makedev(major, minor+i)))
+		if (mknod(name, S_IFBLK | 0600, makedev(major_num, minor_num+i)))
 			perror("mknod");
 		if (chown(name, stb2.st_uid, stb2.st_gid))
 			perror("chown");
@@ -110,10 +110,9 @@ int open_mddev(char *dev, int autof)
 {
 	int mdfd;
 	struct stat stb;
-	int major = MD_MAJOR;
-	int minor = 0;
+	int major_num = MD_MAJOR;
+	int minor_num = 0;
 	int must_remove = 0;
-	struct mdstat_ent *mdlist;
 	int num;
 	struct createinfo *ci = conf_get_create_info();
 	int parts;
@@ -135,41 +134,42 @@ int open_mddev(char *dev, int autof)
 				dev);
 			return -1;
 		}
-		if (autof == 2 && stb.st_mode == 0 && !is_standard(dev, NULL)) {
-			fprintf(stderr, Name ": --auto=yes requires a 'standard' md device name, not %s\n", dev);
-			return -1;
-		}
 		/* check major number is correct */
 		num = -1;
 		std = is_standard(dev, &num);
-		if (std>0) major = get_mdp_major();
+		if (std>0) major_num = get_mdp_major();
 		switch(autof) {
 		case 2: /* only create is_standard names */
 			if (!std && !stb.st_mode) {
-				fprintf(stderr, Name ": --auto=yes requires a 'standard' md device name, not %s\n", dev);
+				fprintf(stderr, Name
+			": %s does not exist and is not a 'standard' name "
+			"so it cannot be created\n", dev);
 				return -1;
 			}
 			break;
 		case 3: /* create md, reject std>0 */
 			if (std > 0) {
-				fprintf(stderr, Name ": that --auto option not compatable with device named %s\n", dev);
+				fprintf(stderr, Name ": that --auto option "
+				"not compatable with device named %s\n", dev);
 				return -1;
 			}
 			break;
 		case 4: /* create mdp, reject std<0 */
 			if (std < 0) {
-				fprintf(stderr, Name ": that --auto option not compatable with device named %s\n", dev);
+				fprintf(stderr, Name ": that --auto option "
+				"not compatable with device named %s\n", dev);
 				return -1;
 			}
+			major_num = get_mdp_major();
 			break;
 		case 5: /* default to md if not standard */
 			break;
 		case 6: /* default to mdp if not standard */
-			if (std == 0) major = get_mdp_major();
+			if (std == 0) major_num = get_mdp_major();
 			break;
 		}
 		/* major is final. num is -1 if not standard */
-		if (stb.st_mode && major(stb.st_rdev) != major)
+		if (stb.st_mode && major(stb.st_rdev) != major_num)
 			must_remove = 1;
 		if (stb.st_mode && !must_remove) {
 			/* looks ok, see if it is available */
@@ -184,7 +184,7 @@ int open_mddev(char *dev, int autof)
 				close(mdfd);
 				return -1;
 			}
-			if (major != MD_MAJOR && parts > 0)
+			if (major_num != MD_MAJOR && parts > 0)
 				make_parts(dev, parts, ci->symlinks);
 			return mdfd;
 		}
@@ -192,56 +192,31 @@ int open_mddev(char *dev, int autof)
 		 * If the device name is in a 'standard' format,
 		 * intuit the minor from that, else
 		 * easiest to read /proc/mdstat, and hunt through for
-		 * an unused number 
+		 * an unused number
 		 */
 		if (num < 0) {
 			/* need to pick an unused number */
-			mdlist = mdstat_read(0, 0);
-			/* Choose a large number.  Start from 127 and search down,
-			 * but if nothing is found, start really big
-			 */
-			for (num = 127 ; num != 128 ; num = num ? num-1 : (1<<22)-1) {
-				struct mdstat_ent *me;
-				int devnum = num;
-				if (major != MD_MAJOR)
-					devnum = -1-num;
+			int num = find_free_devnum(major_num != MD_MAJOR);
 
-				for (me=mdlist; me; me=me->next)
-					if (me->devnum == devnum)
-						break;
-				if (!me) {
-					/* doesn't exist in mdstat.
-					 * make sure it is new to /dev too
-					 */
-					char *dn;
-					if (major != MD_MAJOR)
-						minor = num << MdpMinorShift;
-					else
-						minor = num;
-					dn = map_dev(major,minor, 0);
-					if (dn==NULL || is_standard(dn, NULL)) {
-						/* this number only used by a 'standard' name,
-						 * so it is safe to use
-						 */
-						break;
-					}
-				}
-			}
-		} else if (major == MD_MAJOR)
-			minor = num;
+			if (major_num == MD_MAJOR)
+				minor_num = num;
+			else
+				minor_num = (-1-num) << MdpMinorShift;
+		} else if (major_num == MD_MAJOR)
+			minor_num = num;
 		else
-			minor = num << MdpMinorShift;
+			minor_num = num << MdpMinorShift;
 		/* major and minor have been chosen */
 
 		/* If it was a 'standard' name and it is in-use, then
 		 * the device could already be correct
 		 */
-		if (stb.st_mode && major(stb.st_rdev) == major &&
-		    minor(stb.st_rdev) == minor)
+		if (stb.st_mode && major(stb.st_rdev) == major_num &&
+		    minor(stb.st_rdev) == minor_num)
 			;
 		else {
-			if (major(makedev(major,minor)) != major ||
-			    minor(makedev(major,minor)) != minor) {
+			if (major(makedev(major_num,minor_num)) != major_num ||
+			    minor(makedev(major_num,minor_num)) != minor_num) {
 				fprintf(stderr, Name ": Need newer C library to use more than 4 partitionable md devices, sorry\n");
 				return -1;
 			}
@@ -256,7 +231,7 @@ int open_mddev(char *dev, int autof)
 						perror("chmod /dev/md");
 				}
 			}
-			if (mknod(dev, S_IFBLK|0600, makedev(major, minor))!= 0) {
+			if (mknod(dev, S_IFBLK|0600, makedev(major_num, minor_num))!= 0) {
 				fprintf(stderr, Name ": failed to create %s\n", dev);
 				return -1;
 			}
@@ -275,7 +250,7 @@ int open_mddev(char *dev, int autof)
 			add_dev(dev, &stb, 0, NULL);
 			if (ci->symlinks && strncmp(dev, "/dev/md/", 8) == 0)
 				make_dev_symlink(dev);
-			if (major != MD_MAJOR)
+			if (major_num != MD_MAJOR)
 				make_parts(dev,parts, ci->symlinks);
 		}
 	}
@@ -292,3 +267,82 @@ int open_mddev(char *dev, int autof)
 	return mdfd;
 }
 
+
+int open_mddev_devnum(char *devname, int devnum, char *name,
+		      char *chosen_name, int parts)
+{
+	/* Open the md device with number 'devnum', possibly using 'devname',
+	 * possibly constructing a name with 'name', but in any case, copying
+	 * the name into 'chosen_name'
+	 */
+	int major_num, minor_num;
+	struct stat stb;
+	int i;
+	struct createinfo *ci = conf_get_create_info();
+
+	if (devname)
+		strcpy(chosen_name, devname);
+	else if (name && strchr(name,'/') == NULL) {
+		char *n = strchr(name, ':');
+		if (n) n++; else n = name;
+		if (isdigit(*n) && devnum < 0)
+			sprintf(chosen_name, "/dev/md/d%s", n);
+		else
+			sprintf(chosen_name, "/dev/md/%s", n);
+	} else {
+		if (devnum >= 0)
+			sprintf(chosen_name, "/dev/md%d", devnum);
+		else
+			sprintf(chosen_name, "/dev/md/d%d", -1-devnum);
+	}
+	if (devnum >= 0) {
+		major_num = MD_MAJOR;
+		minor_num = devnum;
+	} else {
+		major_num = get_mdp_major();
+		minor_num = (-1-devnum) << 6;
+	}
+	if (stat(chosen_name, &stb) == 0) {
+		/* It already exists.  Check it is right. */
+		if ( ! S_ISBLK(stb.st_mode) ||
+		     stb.st_rdev != makedev(major_num, minor_num)) {
+			errno = EEXIST;
+			return -1;
+		}
+	} else {
+		/* special case: if --incremental is suggesting a name
+		 * in /dev/md/, we make sure the directory exists.
+		 */
+		if (strncmp(chosen_name, "/dev/md/", 8) == 0) {
+			if (mkdir("/dev/md",0700)==0) {
+				if (chown("/dev/md", ci->uid, ci->gid))
+					perror("chown /dev/md");
+				if (chmod("/dev/md", ci->mode|
+					          ((ci->mode>>2) & 0111)))
+					perror("chmod /dev/md");
+			}
+		}
+
+		if (mknod(chosen_name, S_IFBLK | 0600,
+			  makedev(major_num, minor_num)) != 0) {
+			return -1;
+		}
+		/* FIXME chown/chmod ?? */
+	}
+
+	/* Simple locking to avoid --incr being called for the same
+	 * array multiple times in parallel.
+	 */
+	for (i = 0; i < 25 ; i++) {
+		int fd;
+
+		fd = open(chosen_name, O_RDWR|O_EXCL);
+		if (fd >= 0 || errno != EBUSY) {
+			if (devnum < 0)
+				make_parts(chosen_name, parts, ci->symlinks);
+			return fd;
+		}
+		usleep(200000);
+	}
+	return -1;
+}
