@@ -142,9 +142,9 @@ static unsigned int calc_sb_1_csum(struct mdp_superblock_1 * sb)
 }
 
 #ifndef MDASSEMBLE
-static void examine_super1(void *sbv, char *homehost)
+static void examine_super1(struct supertype *st, char *homehost)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	time_t atime;
 	int d;
 	int faulty;
@@ -152,9 +152,17 @@ static void examine_super1(void *sbv, char *homehost)
 	char *c;
 	int l = homehost ? strlen(homehost) : 0;
 	int layout;
+	unsigned long long sb_offset;
 
 	printf("          Magic : %08x\n", __le32_to_cpu(sb->magic));
-	printf("        Version : %02d\n", 1);
+	printf("        Version : 1");
+	sb_offset = __le64_to_cpu(sb->super_offset);
+	if (sb_offset <= 4)
+		printf(".1\n");
+	else if (sb_offset <= 8)
+		printf(".2\n");
+	else
+		printf(".0\n");
 	printf("    Feature Map : 0x%x\n", __le32_to_cpu(sb->feature_map));
 	printf("     Array UUID : ");
 	for (i=0; i<16; i++) {
@@ -174,7 +182,7 @@ static void examine_super1(void *sbv, char *homehost)
 	printf("     Raid Level : %s\n", c?c:"-unknown-");
 	printf("   Raid Devices : %d\n", __le32_to_cpu(sb->raid_disks));
 	printf("\n");
-	printf("  Used Dev Size : %llu%s\n",
+	printf(" Avail Dev Size : %llu%s\n",
 	       (unsigned long long)__le64_to_cpu(sb->data_size),
 	       human_size(__le64_to_cpu(sb->data_size)<<9));
 	if (__le32_to_cpu(sb->level) >= 0) {
@@ -194,7 +202,7 @@ static void examine_super1(void *sbv, char *homehost)
 			       ddsks*(unsigned long long)__le64_to_cpu(sb->size),
 			       human_size(ddsks*__le64_to_cpu(sb->size)<<9));
 		if (sb->size != sb->data_size)
-			printf("      Used Size : %llu%s\n",
+			printf("  Used Dev Size : %llu%s\n",
 			       (unsigned long long)__le64_to_cpu(sb->size),
 			       human_size(__le64_to_cpu(sb->size)<<9));
 	}
@@ -333,10 +341,11 @@ static void examine_super1(void *sbv, char *homehost)
 }
 
 
-static void brief_examine_super1(void *sbv)
+static void brief_examine_super1(struct supertype *st)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	int i;
+	unsigned long long sb_offset;
 	char *nm;
 	char *c=map_num(pers, __le32_to_cpu(sb->level));
 
@@ -348,9 +357,15 @@ static void brief_examine_super1(void *sbv)
 	else
 		nm = "??";
 
-	printf("ARRAY /dev/md/%s level=%s metadata=1 num-devices=%d UUID=",
-	       nm,
-	       c?c:"-unknown-", __le32_to_cpu(sb->raid_disks));
+	printf("ARRAY /dev/md/%s level=%s ", nm, c?c:"-unknown-");
+	sb_offset = __le64_to_cpu(sb->super_offset);
+	if (sb_offset <= 4)
+		printf("metadata=1.1 ");
+	else if (sb_offset <= 8)
+		printf("metadata=1.2 ");
+	else
+		printf("metadata=1.0 ");
+	printf("num-devices=%d UUID=", __le32_to_cpu(sb->raid_disks));
 	for (i=0; i<16; i++) {
 		if ((i&3)==0 && i != 0) printf(":");
 		printf("%02x", sb->set_uuid[i]);
@@ -360,9 +375,43 @@ static void brief_examine_super1(void *sbv)
 	printf("\n");
 }
 
-static void detail_super1(void *sbv, char *homehost)
+static void export_examine_super1(struct supertype *st)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
+	int i;
+	int len = 32;
+
+	printf("MD_LEVEL=%s\n", map_num(pers, __le32_to_cpu(sb->level)));
+	printf("MD_DEVICES=%d\n", __le32_to_cpu(sb->raid_disks));
+	for (i=0; i<32; i++)
+		if (sb->set_name[i] == '\n' ||
+		    sb->set_name[i] == '\0') {
+			len = i;
+			break;
+		}
+	if (len)
+		printf("MD_NAME=%.*s\n", len, sb->set_name);
+	printf("MD_UUID=");
+	for (i=0; i<16; i++) {
+		if ((i&3)==0 && i != 0) printf(":");
+		printf("%02x", sb->set_uuid[i]);
+	}
+	printf("\n");
+	printf("MD_UPDATE_TIME=%llu\n",
+	       __le64_to_cpu(sb->utime) & 0xFFFFFFFFFFULL);
+	printf("MD_DEV_UUID=");
+	for (i=0; i<16; i++) {
+		if ((i&3)==0 && i != 0) printf(":");
+		printf("%02x", sb->device_uuid[i]);
+	}
+	printf("\n");
+	printf("MD_EVENTS=%llu\n",
+	       (unsigned long long)__le64_to_cpu(sb->events));
+}
+
+static void detail_super1(struct supertype *st, char *homehost)
+{
+	struct mdp_superblock_1 *sb = st->sb;
 	int i;
 	int l = homehost ? strlen(homehost) : 0;
 
@@ -379,9 +428,9 @@ static void detail_super1(void *sbv, char *homehost)
 	printf("\n         Events : %llu\n\n", (unsigned long long)__le64_to_cpu(sb->events));
 }
 
-static void brief_detail_super1(void *sbv)
+static void brief_detail_super1(struct supertype *st)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	int i;
 
 	if (sb->set_name[0])
@@ -393,9 +442,9 @@ static void brief_detail_super1(void *sbv)
 	}
 }
 
-static void export_super1(void *sbv)
+static void export_detail_super1(struct supertype *st)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	int i;
 	int len = 32;
 
@@ -417,9 +466,9 @@ static void export_super1(void *sbv)
 
 #endif
 
-static int match_home1(void *sbv, char *homehost)
+static int match_home1(struct supertype *st, char *homehost)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	int l = homehost ? strlen(homehost) : 0;
 
 	return (l > 0 && l < 32 &&
@@ -427,18 +476,18 @@ static int match_home1(void *sbv, char *homehost)
 		strncmp(sb->set_name, homehost, l) == 0);
 }
 
-static void uuid_from_super1(int uuid[4], void * sbv)
+static void uuid_from_super1(struct supertype *st, int uuid[4])
 {
-	struct mdp_superblock_1 *super = sbv;
+	struct mdp_superblock_1 *super = st->sb;
 	char *cuuid = (char*)uuid;
 	int i;
 	for (i=0; i<16; i++)
 		cuuid[i] = super->set_uuid[i];
 }
 
-static void getinfo_super1(struct mdinfo *info, void *sbv)
+static void getinfo_super1(struct supertype *st, struct mdinfo *info)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	int working = 0;
 	int i;
 	int role;
@@ -507,7 +556,8 @@ static void getinfo_super1(struct mdinfo *info, void *sbv)
 	info->array.working_disks = working;
 }
 
-static int update_super1(struct mdinfo *info, void *sbv, char *update,
+static int update_super1(struct supertype *st, struct mdinfo *info,
+			 char *update,
 			 char *devname, int verbose,
 			 int uuid_set, char *homehost)
 {
@@ -515,7 +565,7 @@ static int update_super1(struct mdinfo *info, void *sbv, char *update,
 	 * For others, the return value is ignored.
 	 */
 	int rv = 0;
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 
 	if (strcmp(update, "force-one")==0) {
 		/* Not enough devices for a working array,
@@ -587,7 +637,7 @@ static int update_super1(struct mdinfo *info, void *sbv, char *update,
 
 		if (__le32_to_cpu(sb->feature_map)&MD_FEATURE_BITMAP_OFFSET) {
 			struct bitmap_super_s *bm;
-			bm = (struct bitmap_super_s*)(sbv+1024);
+			bm = (struct bitmap_super_s*)(st->sb+1024);
 			memcpy(bm->uuid, sb->set_uuid, 16);
 		}
 	}
@@ -620,7 +670,7 @@ static int update_super1(struct mdinfo *info, void *sbv, char *update,
 	    __le64_to_cpu(sb->data_offset)) {
 		/* set data_size to device size less data_offset */
 		struct misc_dev_info *misc = (struct misc_dev_info*)
-			(sbv + 1024 + sizeof(struct bitmap_super_s));
+			(st->sb + 1024 + sizeof(struct bitmap_super_s));
 		printf("Size was %llu\n", (unsigned long long)
 		       __le64_to_cpu(sb->data_size));
 		sb->data_size = __cpu_to_le64(
@@ -635,7 +685,7 @@ static int update_super1(struct mdinfo *info, void *sbv, char *update,
 	return rv;
 }
 
-static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info,
+static int init_super1(struct supertype *st, mdu_array_info_t *info,
 		       unsigned long long size, char *name, char *homehost, int *uuid)
 {
 	struct mdp_superblock_1 *sb = malloc(1024 + sizeof(bitmap_super_t) +
@@ -645,9 +695,9 @@ static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info,
 	char defname[10];
 	memset(sb, 0, 1024);
 
+	st->sb = sb;
 	if (info->major_version == -1) {
 		/* zeroing superblock */
-		*sbp = sb;
 		return 0;
 	}
 
@@ -714,14 +764,13 @@ static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info,
 
 	memset(sb->dev_roles, 0xff, 1024 - sizeof(struct mdp_superblock_1));
 
-	*sbp = sb;
 	return 1;
 }
 
 /* Add a device to the superblock being created */
-static void add_to_super1(void *sbv, mdu_disk_info_t *dk)
+static void add_to_super1(struct supertype *st, mdu_disk_info_t *dk)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	__u16 *rp = sb->dev_roles + dk->number;
 	if ((dk->state & 6) == 6) /* active, sync */
 		*rp = __cpu_to_le16(dk->raid_disk);
@@ -731,11 +780,11 @@ static void add_to_super1(void *sbv, mdu_disk_info_t *dk)
 		*rp = 0xfffe;
 }
 
-static void locate_bitmap1(struct supertype *st, int fd, void *sbv);
+static void locate_bitmap1(struct supertype *st, int fd);
 
-static int store_super1(struct supertype *st, int fd, void *sbv)
+static int store_super1(struct supertype *st, int fd)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	unsigned long long sb_offset;
 	int sbsize;
 	unsigned long long dsize;
@@ -793,7 +842,7 @@ static int store_super1(struct supertype *st, int fd, void *sbv)
 		struct bitmap_super_s *bm = (struct bitmap_super_s*)
 			(((char*)sb)+1024);
 		if (__le32_to_cpu(bm->magic) == BITMAP_MAGIC) {
-			locate_bitmap1(st, fd, sbv);
+			locate_bitmap1(st, fd);
 			if (write(fd, bm, sizeof(*bm)) != sizeof(*bm))
 			    return 5;
 		}
@@ -802,7 +851,7 @@ static int store_super1(struct supertype *st, int fd, void *sbv)
 	return 0;
 }
 
-static int load_super1(struct supertype *st, int fd, void **sbp, char *devname);
+static int load_super1(struct supertype *st, int fd, char *devname);
 
 static unsigned long choose_bm_space(unsigned long devsize)
 {
@@ -817,11 +866,11 @@ static unsigned long choose_bm_space(unsigned long devsize)
 	return 4*2;
 }
 
-static int write_init_super1(struct supertype *st, void *sbv,
+static int write_init_super1(struct supertype *st,
 			     mdu_disk_info_t *dinfo, char *devname)
 {
-	struct mdp_superblock_1 *sb = sbv;
-	void *refsbv = NULL;
+	struct mdp_superblock_1 *sb = st->sb;
+	struct supertype refst;
 	int fd = open(devname, O_RDWR | O_EXCL);
 	int rfd;
 	int rv;
@@ -851,8 +900,10 @@ static int write_init_super1(struct supertype *st, void *sbv,
 	if (rfd >= 0) close(rfd);
 	sb->events = 0;
 
-	if (load_super1(st, fd, &refsbv, NULL)==0) {
-		struct mdp_superblock_1 *refsb = refsbv;
+	refst =*st;
+	refst.sb = NULL;
+	if (load_super1(&refst, fd, NULL)==0) {
+		struct mdp_superblock_1 *refsb = refst.sb;
 
 		memcpy(sb->device_uuid, refsb->device_uuid, 16);
 		if (memcmp(sb->set_uuid, refsb->set_uuid, 16)==0) {
@@ -888,7 +939,7 @@ static int write_init_super1(struct supertype *st, void *sbv,
 	 * for a bitmap.
 	 */
 	array_size = __le64_to_cpu(sb->size);
-	/* work out how much space we left of a bitmap */
+	/* work out how much space we left for a bitmap */
 	bm_space = choose_bm_space(array_size);
 
 	switch(st->minor_version) {
@@ -898,16 +949,22 @@ static int write_init_super1(struct supertype *st, void *sbv,
 		sb_offset &= ~(4*2-1);
 		sb->super_offset = __cpu_to_le64(sb_offset);
 		sb->data_offset = __cpu_to_le64(0);
+		if (sb_offset - bm_space < array_size)
+			bm_space = sb_offset - array_size;
 		sb->data_size = __cpu_to_le64(sb_offset - bm_space);
 		break;
 	case 1:
 		sb->super_offset = __cpu_to_le64(0);
+		if (4*2 + bm_space + __le64_to_cpu(sb->size) > dsize)
+			bm_space = dsize - __le64_to_cpu(sb->size) - 4*2;
 		sb->data_offset = __cpu_to_le64(bm_space + 4*2);
 		sb->data_size = __cpu_to_le64(dsize - bm_space - 4*2);
 		break;
 	case 2:
 		sb_offset = 4*2;
 		sb->super_offset = __cpu_to_le64(4*2);
+		if (4*2 + 4*2 + bm_space + __le64_to_cpu(sb->size) > dsize)
+			bm_space = dsize - __le64_to_cpu(sb->size) - 4*2 - 4*2;
 		sb->data_offset = __cpu_to_le64(4*2 + 4*2 + bm_space);
 		sb->data_size = __cpu_to_le64(dsize - 4*2 - 4*2 - bm_space );
 		break;
@@ -917,17 +974,17 @@ static int write_init_super1(struct supertype *st, void *sbv,
 
 
 	sb->sb_csum = calc_sb_1_csum(sb);
-	rv = store_super1(st, fd, sb);
+	rv = store_super1(st, fd);
 	if (rv)
 		fprintf(stderr, Name ": failed to write superblock to %s\n", devname);
 
 	if (rv == 0 && (__le32_to_cpu(sb->feature_map) & 1))
-		rv = st->ss->write_bitmap(st, fd, sbv);
+		rv = st->ss->write_bitmap(st, fd);
 	close(fd);
 	return rv;
 }
 
-static int compare_super1(void **firstp, void *secondv)
+static int compare_super1(struct supertype *st, struct supertype *tst)
 {
 	/*
 	 * return:
@@ -936,8 +993,8 @@ static int compare_super1(void **firstp, void *secondv)
 	 *  2 wrong uuid
 	 *  3 wrong other info
 	 */
-	struct mdp_superblock_1 *first = *firstp;
-	struct mdp_superblock_1 *second = secondv;
+	struct mdp_superblock_1 *first = st->sb;
+	struct mdp_superblock_1 *second = tst->sb;
 
 	if (second->magic != __cpu_to_le32(MD_SB_MAGIC))
 		return 1;
@@ -949,7 +1006,7 @@ static int compare_super1(void **firstp, void *secondv)
 			       sizeof(struct misc_dev_info));
 		memcpy(first, second, 1024+sizeof(bitmap_super_t) +
 		       sizeof(struct misc_dev_info));
-		*firstp = first;
+		st->sb = first;
 		return 0;
 	}
 	if (memcmp(first->set_uuid, second->set_uuid, 16)!= 0)
@@ -965,7 +1022,9 @@ static int compare_super1(void **firstp, void *secondv)
 	return 0;
 }
 
-static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
+static void free_super1(struct supertype *st);
+
+static int load_super1(struct supertype *st, int fd, char *devname)
 {
 	unsigned long long dsize;
 	unsigned long long sb_offset;
@@ -974,37 +1033,40 @@ static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 	struct bitmap_super_s *bsb;
 	struct misc_dev_info *misc;
 
+	free_super1(st);
 
-	if (st->ss == NULL) {
+	if (st->ss == NULL || st->minor_version == -1) {
 		int bestvers = -1;
+		struct supertype tst;
 		__u64 bestctime = 0;
 		/* guess... choose latest ctime */
-		st->ss = &super1;
-		for (st->minor_version = 0; st->minor_version <= 2 ; st->minor_version++) {
-			switch(load_super1(st, fd, sbp, devname)) {
-			case 0: super = *sbp;
+		tst.ss = &super1;
+		tst.sb = NULL;
+		for (tst.minor_version = 0; tst.minor_version <= 2 ; tst.minor_version++) {
+			switch(load_super1(&tst, fd, devname)) {
+			case 0: super = tst.sb;
 				if (bestvers == -1 ||
 				    bestctime < __le64_to_cpu(super->ctime)) {
-					bestvers = st->minor_version;
+					bestvers = tst.minor_version;
 					bestctime = __le64_to_cpu(super->ctime);
 				}
 				free(super);
-				*sbp = NULL;
+				tst.sb = NULL;
 				break;
-			case 1: st->ss = NULL; return 1; /*bad device */
+			case 1: return 1; /*bad device */
 			case 2: break; /* bad, try next */
 			}
 		}
 		if (bestvers != -1) {
 			int rv;
-			st->minor_version = bestvers;
-			st->ss = &super1;
-			st->max_devs = 384;
-			rv = load_super1(st, fd, sbp, devname);
-			if (rv) st->ss = NULL;
+			tst.minor_version = bestvers;
+			tst.ss = &super1;
+			tst.max_devs = 384;
+			rv = load_super1(&tst, fd, devname);
+			if (rv == 0)
+				*st = tst;
 			return rv;
 		}
-		st->ss = NULL;
 		return 2;
 	}
 	if (!get_dev_size(fd, devname, &dsize))
@@ -1085,7 +1147,7 @@ static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 		free(super);
 		return 2;
 	}
-	*sbp = super;
+	st->sb = super;
 
 	bsb = (struct bitmap_super_s *)(((char*)super)+1024);
 
@@ -1099,12 +1161,12 @@ static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 	 * valid.  If it doesn't clear the bit.  An --assemble --force
 	 * should get that written out.
 	 */
-	locate_bitmap1(st, fd, super);
+	locate_bitmap1(st, fd);
 	if (read(fd, ((char*)super)+1024, sizeof(struct bitmap_super_s))
 	    != sizeof(struct bitmap_super_s))
 		goto no_bitmap;
 
-	uuid_from_super1(uuid, super);
+	uuid_from_super1(st, uuid);
 	if (__le32_to_cpu(bsb->magic) != BITMAP_MAGIC ||
 	    memcmp(bsb->uuid, uuid, 16) != 0)
 		goto no_bitmap;
@@ -1123,9 +1185,8 @@ static struct supertype *match_metadata_desc1(char *arg)
 
 	st->ss = &super1;
 	st->max_devs = 384;
-	if (strcmp(arg, "1") == 0 ||
-	    strcmp(arg, "1.0") == 0 ||
-	    strcmp(arg, "default/large") == 0) {
+	st->sb = NULL;
+	if (strcmp(arg, "1.0") == 0) {
 		st->minor_version = 0;
 		return st;
 	}
@@ -1137,6 +1198,11 @@ static struct supertype *match_metadata_desc1(char *arg)
 		st->minor_version = 2;
 		return st;
 	}
+	if (strcmp(arg, "1") == 0 ||
+	    strcmp(arg, "default/large") == 0) {
+		st->minor_version = -1;
+		return st;
+	}
 
 	free(st);
 	return NULL;
@@ -1148,12 +1214,26 @@ static struct supertype *match_metadata_desc1(char *arg)
  */
 static __u64 avail_size1(struct supertype *st, __u64 devsize)
 {
+	struct mdp_superblock_1 *super = st->sb;
 	if (devsize < 24)
 		return 0;
 
-	devsize -= choose_bm_space(devsize);
+	if (super == NULL)
+		/* creating:  allow suitable space for bitmap */
+		devsize -= choose_bm_space(devsize);
+#ifndef MDASSEMBLE
+	else if (__le32_to_cpu(super->feature_map)&MD_FEATURE_BITMAP_OFFSET) {
+		/* hot-add. allow for actual size of bitmap */
+		struct bitmap_super_s *bsb;
+		bsb = (struct bitmap_super_s *)(((char*)super)+1024);
+		devsize -= bitmap_sectors(bsb);
+	}
+#endif
 
 	switch(st->minor_version) {
+	case -1: /* no specified.  Now time to set default */
+		st->minor_version = 0;
+		/* FALL THROUGH */
 	case 0:
 		/* at end */
 		return ((devsize - 8*2 ) & ~(4*2-1));
@@ -1168,7 +1248,7 @@ static __u64 avail_size1(struct supertype *st, __u64 devsize)
 }
 
 static int
-add_internal_bitmap1(struct supertype *st, void *sbv,
+add_internal_bitmap1(struct supertype *st,
 		     int *chunkp, int delay, int write_behind,
 		     unsigned long long size,
 		     int may_change, int major)
@@ -1190,7 +1270,7 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 	long offset;
 	int chunk = *chunkp;
 	int room = 0;
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	bitmap_super_t *bms = (bitmap_super_t*)(((char*)sb) + 1024);
 
 	switch(st->minor_version) {
@@ -1275,7 +1355,7 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 	memset(bms, 0, sizeof(*bms));
 	bms->magic = __cpu_to_le32(BITMAP_MAGIC);
 	bms->version = __cpu_to_le32(major);
-	uuid_from_super1((int*)bms->uuid, sb);
+	uuid_from_super1(st, (int*)bms->uuid);
 	bms->chunksize = __cpu_to_le32(chunk);
 	bms->daemon_sleep = __cpu_to_le32(delay);
 	bms->sync_size = __cpu_to_le64(size);
@@ -1286,18 +1366,18 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 }
 
 
-static void locate_bitmap1(struct supertype *st, int fd, void *sbv)
+static void locate_bitmap1(struct supertype *st, int fd)
 {
 	unsigned long long offset;
 	struct mdp_superblock_1 *sb;
 	int mustfree = 0;
 
-	if (!sbv) {
-		if (st->ss->load_super(st, fd, &sbv, NULL))
+	if (!st->sb) {
+		if (st->ss->load_super(st, fd, NULL))
 			return; /* no error I hope... */
 		mustfree = 1;
 	}
-	sb = sbv;
+	sb = st->sb;
 
 	offset = __le64_to_cpu(sb->super_offset);
 	offset += (int32_t) __le32_to_cpu(sb->bitmap_offset);
@@ -1306,16 +1386,16 @@ static void locate_bitmap1(struct supertype *st, int fd, void *sbv)
 	lseek64(fd, offset<<9, 0);
 }
 
-static int write_bitmap1(struct supertype *st, int fd, void *sbv)
+static int write_bitmap1(struct supertype *st, int fd)
 {
-	struct mdp_superblock_1 *sb = sbv;
+	struct mdp_superblock_1 *sb = st->sb;
 	bitmap_super_t *bms = (bitmap_super_t*)(((char*)sb)+1024);
 	int rv = 0;
 
 	int towrite, n;
 	char buf[4096];
 
-	locate_bitmap1(st, fd, sbv);
+	locate_bitmap1(st, fd);
 
 	if (write(fd, ((char*)sb)+1024, sizeof(bitmap_super_t)) !=
 	    sizeof(bitmap_super_t))
@@ -1340,13 +1420,21 @@ static int write_bitmap1(struct supertype *st, int fd, void *sbv)
 	return rv;
 }
 
+static void free_super1(struct supertype *st)
+{
+	if (st->sb)
+		free(st->sb);
+	st->sb = NULL;
+}
+
 struct superswitch super1 = {
 #ifndef MDASSEMBLE
 	.examine_super = examine_super1,
 	.brief_examine_super = brief_examine_super1,
+	.export_examine_super = export_examine_super1,
 	.detail_super = detail_super1,
 	.brief_detail_super = brief_detail_super1,
-	.export_super = export_super1,
+	.export_detail_super = export_detail_super1,
 #endif
 	.match_home = match_home1,
 	.uuid_from_super = uuid_from_super1,
@@ -1363,6 +1451,7 @@ struct superswitch super1 = {
 	.add_internal_bitmap = add_internal_bitmap1,
 	.locate_bitmap = locate_bitmap1,
 	.write_bitmap = write_bitmap1,
+	.free_super = free_super1,
 	.major = 1,
 #if __BYTE_ORDER == BIG_ENDIAN
 	.swapuuid = 0,
